unit MainUnit;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, ExtCtrls;

type
  TForm1 = class(TForm)
    Panel1: TPanel;
    Label1: TLabel;
    DeviceCombo: TComboBox;
    Panel4: TPanel;
    Label4: TLabel;
    FileCombo: TComboBox;
    Panel3: TPanel;
    Label2: TLabel;
    Label3: TLabel;
    MillisecondCombo: TComboBox;
    DepthCombo: TComboBox;
    Panel2: TPanel;
    GoButton: TButton;
    StopButton: TButton;
    LogMemo: TMemo;
    procedure FormCreate(Sender: TObject);
    procedure GoButtonClick(Sender: TObject);
    procedure StopButtonClick(Sender: TObject);
  protected
    procedure IdleQueue(Sender: TObject; var Done: Boolean);
    { Private declarations }
  public
    { Public declarations }
  end;

var
  Form1: TForm1;

implementation

{$R *.DFM}

uses
  AIOUSB;

function InStrC(C: Char; S: AnsiString; StartIndex: Integer = 1): LongWord;
var
  I: Integer;

begin
  for I := StartIndex to Length(S) do
    if S[I] = C then begin
      Result := I;
      Exit;
    end
  ;

  Result := 0;
end;

procedure LoadDefaultFont(IntoFont: TFont);
var
  FontSize: Integer;
  NCM: TNonClientMetrics;

begin
  NCM.cbSize := SizeOf(NCM);
  if SystemParametersInfo(SPI_GETNONCLIENTMETRICS, SizeOf(NCM), @NCM, 0) then begin
    FontSize := Round(-3/4 * NCM.lfMessageFont.lfHeight);
    IntoFont.Name := NCM.lfMessageFont.lfFaceName;
  end
  else begin
    FontSize := Abs(IntoFont.Size);
    IntoFont.Name := 'Tahoma';
  end;
  IntoFont.PixelsPerInch := 72;
  IntoFont.Size := FontSize;
end;

function HWave_Start(DeviceIndex: LongWord; Milliseconds, BufferDepth: LongInt): LongWord; cdecl; external 'HWave.dll';
function HWave_Stop(DeviceIndex: LongWord): LongWord; cdecl; external 'HWave.dll';
function HWave_Queue(DeviceIndex: LongWord; pData: PWord): LongWord; cdecl; external 'HWave.dll';

const
  NoDeviceStr = '';

type
  TDACEntry = packed record
    Channel: Word;
    Value: Word;
  end;
  PDACEntry = ^TDACEntry;

  TDACRow = array[$0..$F] of Word;

var
  MidScale: array of TDACEntry;
  DeviceIndex: LongWord;
  DACData: array of TDACRow;
  iWaveform: Integer;
  BusyCt, MostBusyCt, MostQueueCt, FirstQueueCt: Integer;

procedure TForm1.FormCreate(Sender: TObject);
var
  iDI, iDAC: Integer;
  DeviceMask: LongWord;
  Status, PID, NomSize: LongWord;
  Nom: AnsiString;
  Findee: TSearchRec;

begin
  IsMultiThread := True;
  LoadDefaultFont(Font);
  DeviceCombo.Align := alTop;
  FileCombo.Align := alTop;

  //Prep array to set all DACs to mid-scale, via DACMultiDirect.
  SetLength(MidScale, 16);
  for iDAC := 0 to 15 do MidScale[iDAC].Channel := iDAC;
  for iDAC := 0 to 15 do MidScale[iDAC].Value := $8000;

  //Stuff device combo.
  DeviceMask := GetDevices;
  for iDI := 0 to 31 do begin
    if (DeviceMask and (1 shl iDI)) = 0 then begin
      //No device at this index, put a gap at this slot so ItemIndex = DeviceIndex.
      Nom := NoDeviceStr;
    end
    else begin
      NomSize := 256;
      SetString(Nom, nil, NomSize);
      Status := QueryDeviceInfo(iDI, @PID, @NomSize, @Nom[1], nil, nil);
      if Status <> ERROR_SUCCESS then begin
        Nom := 'Error ' + IntToStr(Status) + ' From QueryDeviceInfo';
      end
      else begin
        SetLength(Nom, NomSize);
        Nom := Nom + ' [' + IntToHex(PID, 4) + ']';
      end;
    end;

    DeviceCombo.Items.Add(Nom);
  end;

  //There will normally be a bunch of gaps at the end of the list, but it's only
  //useful to keep gaps in the middle. So get rid of gaps from the end.
  for iDI := DeviceCombo.Items.Count-1 downto 0 do
    if DeviceCombo.Items[iDI] = NoDeviceStr then
      DeviceCombo.Items.Delete(iDI)
    else
      Break
  ;

  if DeviceCombo.Items.Count > 0 then begin
    //If we have any actual devices, select the first and get ready to go.
    DeviceCombo.ItemIndex := 0;
    GoButton.Enabled := True;
    ActiveControl := GoButton;
  end
  else begin
    //Else show just a gap.
    DeviceCombo.Items.Add(NoDeviceStr);
    DeviceCombo.ItemIndex := 0;
  end;

  //Stuff file combo.
  if FindFirst('*.tsv', faAnyFile and not faDirectory, Findee) = ERROR_SUCCESS then repeat
    FileCombo.Items.Add(Findee.Name);
  until FindNext(Findee) <> ERROR_SUCCESS;
  if FileCombo.Items.Count <> 0 then FileCombo.ItemIndex := 0;

  with MillisecondCombo do ItemIndex := Items.IndexOf(Text);
  with DepthCombo do ItemIndex := Items.IndexOf(Text);
end;

procedure TForm1.GoButtonClick(Sender: TObject);
var
  Status: LongWord;
  Milliseconds, QueueDepth: LongInt;
  WaveformList, EntryList: TStringList;
  iEntry, iChannel, iFirstChar, iLastChar: Integer;
  ThisRowStr, ThisCellStr: String;
  ThisDACData: TDACRow;

begin
  DeviceIndex := DeviceCombo.ItemIndex;
  if DeviceIndex >= $80000000 then Exit;
  if DeviceCombo.Items[DeviceIndex] = NoDeviceStr then Exit;

  //Load the waveform.
  for iChannel := $0 to $F do ThisDACData[iChannel] := $8000;
  WaveformList := TStringList.Create;
  EntryList := TStringList.Create;
  try
    WaveformList.LoadFromFile(FileCombo.Text);
    SetLength(DACData, 0);
    SetLength(DACData, WaveformList.Count);
    for iEntry := 0 to High(DACData) do begin
      ThisRowStr := WaveformList[iEntry];
      iChannel := 0;
      iFirstChar := 1;
      repeat
        iLastChar := InStrC(#9, ThisRowStr, iFirstChar);
        if iLastChar = 0 then iLastChar := Length(ThisRowStr)+1;

        ThisCellStr := Copy(ThisRowStr, iFirstChar, iLastChar-iFirstChar);
        if ThisCellStr <> '' then
          ThisDACData[iChannel] := Round($FFFF/5 * StrToFloat(ThisCellStr))
        ;
        iFirstChar := iLastChar + 1;
        Inc(iChannel);
      until iFirstChar > Length(ThisRowStr);

      DACData[iEntry] := ThisDACData;
    end;
  finally
    FreeAndNil(WaveformList);
    FreeAndNil(EntryList);
  end;

  Milliseconds := StrToInt(MillisecondCombo.Text);
  QueueDepth := StrToInt(DepthCombo.Text);

  //Set the DACs to mid-scale in prep for sine wave, and turn them on.
  DACMultiDirect(DeviceIndex, @MidScale[0], Length(MidScale));
  DACSetBoardRange(DeviceIndex, 0);

  iWaveform := 0;
  BusyCt := 0;
  MostBusyCt := 0;
  MostQueueCt := 0;
  FirstQueueCt := 0;

  //Start the wave sequence. It will simply block until data is queued.
  Status := HWave_Start(DeviceIndex, Milliseconds, QueueDepth);
  if Status <> ERROR_SUCCESS then begin
    LogMemo.Lines.Add('Error ' + IntToStr(Status) + ' from HWave_Start.');
    Exit;
  end;

  LogMemo.Lines.Add('Started.');

  //We queue in a fast-return idle handler. It's simple, but kinda dumb.
  Application.OnIdle := IdleQueue;

  DeviceCombo.Enabled := False;
  MillisecondCombo.Enabled := False;
  DepthCombo.Enabled := False;

  StopButton.Enabled := True;
  GoButton.Enabled := False;

  StopButton.SetFocus;
end;

procedure TForm1.StopButtonClick(Sender: TObject);
var
  Status: LongWord;

begin
  GoButton.Enabled := True;
  StopButton.Enabled := False;

  DeviceCombo.Enabled := True;
  MillisecondCombo.Enabled := True;
  DepthCombo.Enabled := True;

  Application.OnIdle := nil;

  //Stop the wave sequence. Note that this will discard any queued DAC data. If this
  //call returns an error, then the wave sequence isn't running anyways, so we
  //proceed even if it errors.
  Status := HWave_Stop(DeviceIndex);
  if Status <> ERROR_SUCCESS then begin
    LogMemo.Lines.Add('Error ' + IntToStr(Status) + ' from HWave_Stop.');
  end;

  //Reset the DACs to mid-scale.
  DACMultiDirect(DeviceIndex, @MidScale[0], Length(MidScale));
  
  LogMemo.Lines.Add('Stopped.');

  LogMemo.Lines.Add('Most Contiguous Busies: ' + IntToStr(MostBusyCt));
  LogMemo.Lines.Add('First Contiguous Queues: ' + IntToStr(FirstQueueCt));
  LogMemo.Lines.Add('Most Contiguous Queues Thereafter: ' + IntToStr(MostQueueCt));

  GoButton.SetFocus;
end;

procedure TForm1.IdleQueue(Sender: TObject; var Done: Boolean);
var
  Status: LongWord;
  //iDAC: Integer;
  QueueCt: Integer;
  //BaseTheta: Extended;

begin
  try

    QueueCt := 0;
    repeat
      {
      //Generate the DAC data.
      BaseTheta := 2*Pi/19 * BaseClock;
      for iDAC := 0 to High(DACData) do
        DACData[iDAC] := Round($FFFF/2 * (1 + Sin(BaseTheta * (1 + iDAC))))
      ;
      }
      //Queue the DAC data.
      Status := HWave_Queue(DeviceIndex, @DACData[iWaveform, 0]);
      if Status <> ERROR_SUCCESS then Break;

      Inc(QueueCt);

      //If queue succeeded, then advance the clock for the next DAC data.
      iWaveform := (iWaveform + 1) mod Length(DACData); //BaseClock := (BaseClock + 1) mod 19;
    until False;

    //Update timing notes.
    Inc(BusyCt);
    if QueueCt <> 0 then begin
      if MostBusyCt < BusyCt then MostBusyCt := BusyCt;
      BusyCt := 0;
      if FirstQueueCt = 0 then
        FirstQueueCt := QueueCt
      else if MostQueueCt < QueueCt then
        MostQueueCt := QueueCt
      ;
    end;

    Done := False;
  except
    Application.OnIdle := nil;
    raise;
  end;
end;

end.
